using System;
using System.Net.Sockets;
using System.Reflection;
using System.Threading;
using Microsoft.SPOT;

namespace Kaushik.Spot.Library
{
    /// <summary>
    /// This class provides foundation of SPOT Services and provides all necessary functions to deal with SPOT Services.
    /// </summary>
    public static class SpotServices
    {
        #region Member Variables
        private static readonly short _maxServices;
        private static object[] _services;
        private static short _currentServices;
        private static short _port;
        private static bool _serverRunning;
        private static ICryptographicServiceProvider _cipher = null;
        #endregion Member Variables

        #region SpotServices
        /// <summary>
        /// Constructor
        /// </summary>
        static SpotServices()
        {
            _maxServices = 5;
            _services = new object[_maxServices];
            _currentServices = 0;
            _serverRunning = false;
        }
        #endregion SpotServices

        #region RegisterService
        /// <summary>
        /// This function is used to register a SPOT Service
        /// </summary>
        /// <param name="service">Object of SPOT Service Class</param>
        public static void RegisterService(object service)
        {
            if (service != null)
            {
                if (_currentServices < _maxServices)
                {
                    lock (_services)
                    {
                        _services[_currentServices] = service;
                    }
                    _currentServices++;
                }
                else
                {
                    throw new Exception("Maximum number (which is " + _maxServices.ToString() + ") of services already registered!");
                }
            }
            else
            {
                throw new NullReferenceException();
            }
        }
        #endregion RegisterService

        #region ExecuteCommand
        /// <summary>
        /// This is function is used to invoke a SPOT Service Method.
        /// </summary>
        /// <param name="command">Object of ServiceCommand</param>
        /// <returns>Object of ServiceResult that has return value and any exception details</returns>
        public static ServiceResult ExecuteCommand(ServiceCommand command)
        {
            ServiceResult result = new ServiceResult();
            bool serviceFound = false;

            try
            {
                if (command != null)
                {
                    for (int counter = 0; counter < _currentServices; counter++)
                    {
                        if (Array.IndexOf(_services[counter].GetType().GetInterfaces(), Type.GetType(command.ServiceName)) != -1)
                        {
                            serviceFound = true;

                            Type serviceType = _services[counter].GetType();
                            MethodInfo method = serviceType.GetMethod(command.Method);
                            if (method != null)
                            {
                                result.Result = method.Invoke(_services[counter], command.Parameters);
                            }
                            else
                            {
                                throw new NotSupportedException("Requested method not supported.");
                            }
                            break;
                        }
                    }

                    Debug.GC(true);

                    if (!serviceFound)
                    {
                        throw new NotSupportedException("Requested service not supported.");
                    }
                }
                else
                {
                    throw new Exception("Service call does not have Command information.");
                }
            }
            catch (Exception error)
            {
                result.Error = error.Message + "\n" + "Error Details:" + error.ToString();
            }

            return result;
        }

        /// <summary>
        /// This is function is used to invoke a SPOT Service Method using Serialized ServiceCommand object.
        /// This function also calls Encrypt and Decrypt if required.
        /// </summary>
        /// <param name="commandText">Serialized ServiceCommand object</param>
        /// <returns>Serialized ServiceResult object</returns>
        public static string ExecuteCommand(string commandText)
        {
            ServiceCommand command = new ServiceCommand();
            string resultText = null;

            if (_cipher == null)
            {
                ServiceResult result = null;
                try
                {
                    command.Deserialize(commandText);
                    result = ExecuteCommand(command);
                    resultText = result.Serialize();
                }
                catch (Exception error)
                {
                    result = new ServiceResult();
                    result.Error = error.Message + "\n" + error.ToString();
                    resultText = result.Serialize();
                }
            }
            else
            {
                ServiceResult result = null;
                try
                {
                    command.Deserialize(_cipher.Decrypt(commandText));
                    result = ExecuteCommand(command);
                    resultText = _cipher.Encrypt(result.Serialize());
                }
                catch (Exception error)
                {
                    result = new ServiceResult();
                    result.Error = error.Message + "\n" + error.ToString();
                    resultText = _cipher.Encrypt(result.Serialize());
                }
            }


            return resultText;
        }
        #endregion ExecuteCommand

        #region StartServer
        /// <summary>
        /// This function is used to start SPOT Services on a new thread. It will make sure if the Server is already running or not.
        /// </summary>
        /// <param name="port">Port on which services should be avalable, default 8080</param>
        /// <param name="cipher">Cipher program that can be used to protect calls over network.</param>
        public static void StartServer(short port = 8080, ICryptographicServiceProvider cipher = null)
        {
            if (!_serverRunning)
            {
                _port = port;
                _cipher = cipher;
                Thread serverThread = new Thread(startServer);
                serverThread.Start();
                _serverRunning = true;
            }
        }

        /// <summary>
        /// This function is used to start SPOT Services
        /// </summary>
        private static void startServer()
        {
            try
            {
                ISpotSocketServer server = SpotSocketFactory.GetServerProvider(SpotSocketProvider.Tcp, _port, 5120, 50000000, 0, 500);

                string data;

                using (Socket listner = server.CreatetListner())
                {
                    while (true)
                    {
                        using (Socket socket = server.Accept())
                        {
                            try
                            {
                                server.Receive(out data);

                                string result = SpotServices.ExecuteCommand(data);

                                server.Send(result);

                                socket.Close();
                            }
                            catch (Exception ex)
                            {
                                string message = ex.Message;
                                Debug.Print(ex.Message);
                            }
                        }

                        Debug.GC(true);
                    }
                }
            }
            catch (Exception ex)
            {
                string message = ex.Message;
                Debug.Print(ex.Message);
            }
        }
        #endregion StartServer
    }
}
